/*
* Copyright 2003(c) Theo Pavlidis
* See pack4png.h for a detailed notice on the use of the software
*/
	
// HistAnalArray.cpp: implementation of the CHistAnalArray class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "HistAnalArray.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHistAnalArray::CHistAnalArray()
{
	m_hi   = NULL;
	m_hh   = NULL;
	m_zone = NULL;
}

CHistAnalArray::~CHistAnalArray()
{
	if(m_hi != NULL)	delete[] m_hi;
	if(m_hh != NULL)	delete[] m_hh;
	if(m_zone != NULL)	delete[] m_zone;
}

BOOL CHistAnalArray::Initialize(long hist[], int levels, int strategy, BOOL answer)
{
	if(levels > 256) return FALSE;
	int i, total;
	for(i=0, total=0; i<levels; i++) total += hist[i];
	if(total<1) return FALSE;
	m_lvl = levels;

	m_hi = new long[levels];		if(m_hi==NULL) return FALSE;
	m_hh = new double[levels];		if(m_hh==NULL) return FALSE;
	m_zone = new hZone[levels/2];	if(m_zone==NULL) return FALSE;
	m_nz = 0;

	for(i=0; i<m_lvl; i++) m_hi[i] = hist[i];	// save unscaled values
	for(i=0; i<m_lvl; i++) m_hh[i] = (double)hist[i]/total;
	for(i=0; i<m_lvl; i++) m_hh[i] *= 100;	// to express it as % - makes displayed values easier to track

	// convert "strategy" to paramers
	if(strategy < 0) strategy = -strategy;
	if(strategy > 10) {	// assumes MR_LVL 9 or less
		strategy = strategy%10;
		m_parcelZones =  TRUE;
	}
	else m_parcelZones = FALSE;
	if(strategy<1) strategy = 1;
	m_min_level = strategy;

//	if(answer==TRUE) {
		TRACE("\n**** p4p ****\n");
		TRACE("qual/minlvl=%d parcelZones=%d\n", m_min_level, m_parcelZones);
//	}
	return TRUE;
}

void CHistAnalArray::ListZones(CString note)
{
	TRACE("%s\n", note.GetBuffer(0));
	CString tmp;
	for(int i=0; i<m_nz; i++) {
		m_zone[i].display(&tmp); TRACE("%s\n", tmp.GetBuffer(0));
	}
}

//	eliminate empty zones
void CHistAnalArray::ZoneCleanUp()
{
	int j = m_nz;
	for(int i=0; i<j; i++) {
		if(m_zone[i].i1 == m_zone[i].i2 || m_zone[i].Peak[0]<0) {	// negative peak test added 6/3/03
			for(int jj=i; jj<j-1; jj++) m_zone[jj+1].copyto(&(m_zone[jj]));
			i--;	// re-examine new entry
			j--;
		}
	}
	m_nz = j;
}

void CHistAnalArray::EliminateSparseZones(int force)
{
	int i, jj;
	for(i=1; i<m_nz-1; i++) {
		if(m_zone[i].Peak[0] <1) {	// both Peak[0] and Peak[1] are 0 in sparse zones
			if(force<1 && m_zone[i].sum>0) continue;
			jj = (m_zone[i].i1 + m_zone[i].i2)/2;
			m_zone[i-1].i2 = jj;
			m_zone[i+1].i1 = jj;
			m_zone[i].i1 = m_zone[i].i2 = jj;
			//	no point in recomputing stats, this is normally the final step before calling Assign()
		}
	}
	ZoneCleanUp();
}

void CHistAnalArray::AbsorbEndZone (int n)
{
	if(m_nz<2) return;	// there may be only 2 zones and absorption make senes
	int i_outside = n==LEFT_Z? 0: m_nz-1;	int i_first  = n==LEFT_Z? 0: m_nz-2;
	int i_inside  = n==LEFT_Z? 1: m_nz-2;	int i_second = n==LEFT_Z? 1: m_nz-1;

	if( m_zone[i_inside].Peak[0]>0 && (m_zone[i_outside].sum < H_RARE_VALUES) ) { // no need to check Peak[1]
		m_zone[i_first].i2  = m_zone[i_first].i1;	// collapse first zone
		m_zone[i_second].i1 = m_zone[i_first].i1;
		m_zone[i_second].Stats(m_hh);
		ZoneCleanUp();
		MakeZoneSymmetric(n==LEFT_Z);	// 1 => first
	}
}

BOOL CHistAnalArray::MakeZoneSymmetric(BOOL first)
{
	const int tol = 3;	// why ?? to avoid silly work
	const int safe = 2; // reduce an interval by more than a half its length
	int diff;
	int ic = first? 0: m_nz-1;
	int d1 = m_zone[ic].imax - m_zone[ic].i1;
	int d2 = m_zone[ic].i2-1 - m_zone[ic].imax;
	if(first) {
		diff = d1 - d2;
		if(diff > tol) {
			if( diff > (m_zone[ic+1].i2 - m_zone[ic+1].i1)/safe ) return FALSE;
			m_zone[ic].i2   += (diff - tol);
			m_zone[ic+1].i1 += (diff - tol);
			m_zone[ic].Stats(m_hh);
			m_zone[ic+1].Stats(m_hh); 
		}
	}
	else {
		diff = d2 - d1;
		if(diff > tol) {
			if( diff > (m_zone[ic-1].i2 - m_zone[ic-1].i1)/safe ) return FALSE;
			m_zone[ic].i1   -= (diff - tol);
			m_zone[ic-1].i2 -= (diff - tol);
			m_zone[ic].Stats(m_hh);
			m_zone[ic-1].Stats(m_hh); 
		}
	}
	return TRUE;
}

//	save final results
int CHistAnalArray::Assign(int pxmap[], int mode, RECT z[])
{
	int i, j, value, nv, rm, tm;
	m_err = 0;
	//CString sdbg;
	if(pxmap == NULL) return m_nz;
	for(i=0; i<m_nz; i++) {
		value = m_zone[i].SpreadOut()?  (int)(m_zone[i].mean+0.5) : m_zone[i].imax;
		if(mode>0 && value>0) {	// modify value to be multiple of mode
			nv = value/mode;
			rm = value - nv*mode;
			if(rm <= mode/4) tm = nv*mode;
			else {
				if(rm >= (3*mode)/4) tm = (nv+1)*mode;
				else tm = ((2*nv+1)*mode)/2;
			}
			//sdbg.Format("zone %d: %d => %d, %d => %d", i, value, nv*mode, rm, tm);
			//AfxMessageBox(sdbg);
			value = tm;
		}
		if(z != NULL) {
			z[i].left = m_zone[i].i1;
			z[i].right = m_zone[i].i2;
			z[i].top = value;
			z[i].bottom = m_zone[i].Peak[0] + m_zone[i].Peak[1];	// check how it affects use in PixUT
		}
		for(j=m_zone[i].i1; j<m_zone[i].i2; j++) {
			pxmap[j] = value;
			m_err += (j-value)*(j-value)*m_hh[j];
		}
	}
	m_err = m_err/100;	// histogram is normalized to 100
	return m_nz;		// n_nz is private membe rof class
}

//////////////////////////////////////////////////////////////////////
////////////////////  LAST BUT NOT LEAST  ////////////////////////////
//	The FormZones(int, int) procedure is the core of the pack4png algorithm //
//////////////////////////////////////////////////////////////////////

//	Separate file from main class file during development

#define MAX_C	24

typedef struct _znode {
	int parent;
	int y, xs, xe;
	int nchild;
	int children[MAX_C];
} znode;

#define MX_NO_CROSS	256	// safely on the high side. Cannot possibly be greater than half the pixel levels

static double scav[MR_LVL] = { 2.5, 2.0, 1.5, 1.0, 0.8, 0.6, 0.4, 0.2, 0.1 };

int CHistAnalArray::FormZones(int forced, int color)
{
	BOOL above;
	int i, j;
	int k, n;
	int LevelNumber = MR_LVL;

	// find true range
	int ileft, iright;
	for(ileft=0; m_hi[ileft]==0 && ileft<m_lvl; ileft++);
	for(iright=m_lvl-1; m_hi[iright]==0 && iright>=0; iright--);	iright++;	// to set it as limit

	double sum = 0;
	for(i=0; i<m_lvl; i++) sum += m_hh[i];	// even if normalised - we do not know normalization scale
	double avg = sum/m_lvl;
	double hmax = m_hh[0];
	for(i=1; i<m_lvl; i++) if(hmax < m_hh[i]) hmax = m_hh[i];
	if(hmax < 1.0) return 0;

	double bound;

	znode * N = new znode[LevelNumber*m_lvl];	// about double the worst case 
	int nz = 0;			// for nodes
	int lastnz[MR_LVL];	// for nodes

	int		div[MX_NO_CROSS];	// places where the histogram crosses a threshold
	BYTE	mrk[MX_NO_CROSS];	// used mainly for debugging but also for eliminating noisy intervals
	double	amp[MX_NO_CROSS];	// amplitude around a crossing
	double hbot, htop;
	int len;
	for(j=0; j<LevelNumber; j++) {
		bound = avg*scav[LevelNumber-1-j];	// start from lowest level
		above = (m_hh[ileft] < bound) ? FALSE: TRUE;
		len = 0;
		div[len] = ileft;
		mrk[len] = above==TRUE? 'a': 'b';
		for(i = ileft+1; i < iright; i++) {
			if(m_hh[i] <  bound && above == FALSE) 	continue;
			if(m_hh[i] >= bound && above == TRUE) continue;
			len++;
			div[len] = i;
			mrk[len] = above==TRUE? 'b': 'a';	// about to be negated
			above = 1 - above;
		}
		len++;
		div[len] = iright;
		mrk[len] = ' ';

		//	Find amplitude around each dividing point
		htop = hbot = bound;
		for(i=0; i<len; i++) {
			if(mrk[i]=='b') {
				hbot = bound;
				for(k=div[i]; k<div[i+1]; k++) if(m_hh[k]<hbot) hbot = m_hh[k];
			}
			else {
				htop = bound;
				for(k=div[i]; k<div[i+1]; k++) if(m_hh[k]>htop) htop = m_hh[k];
			}
			amp[i] = htop - hbot;
		}
		amp[0] = amp[len] = 0;	// we never use them

		// mark intervals of small length
		if(len > 4) {	// a cludge for avoiding elminating everything! trouble occurs for 3 
			for(i=1; i<len-1; i++) {
				if(div[i+1]-div[i-1] > H_MIN_INTERVAL) continue;
				if(amp[i]>H_LARGE_AMPLITUDE) continue;
				mrk[i-1] += 'A'-'a';
				mrk[i]	 += 'A'-'a';
				i++;	// skip pair
			}
			// CLEAN UP (works well - 4/8/03)
			for(i=0; i<len; i++) {
				if(mrk[i] != 'A' && mrk[i] != 'B') continue;
				k = i;
				while(mrk[k]=='A' || mrk[k]=='B') k++;
				// remove all dividers from i to k-1
				len -= (k-i);
				for(n=i; n<=len; n++) {
					div[n] = div[n+k-i];
					mrk[n] = mrk[n+k-i];
				}
			}
		}

		// construct nodes of the tree
		for(i=0; i<len; i++) {
			if(mrk[i]=='a') {
				N[nz].y = j;	N[nz].xs = div[i];	N[nz].xe = div[i+1];
				N[nz].nchild = 0;
				N[nz].parent = -1;	//NULL;
				nz++;
			}
		}
		lastnz[j] = nz;
	}

	//	link nodes of the tree
	for(i=0; i<nz && N[i].y==0; i++);
	lastnz[0] = i;
	for(j=1; j<LevelNumber; j++) {
		n = j<2? 0: lastnz[j-2];
		for(; i<nz && N[i].y==j; i++) {
			for(k=n; k<lastnz[j-1]; k++) {
				ASSERT(N[k].y==j-1);
				if(N[k].xs <= N[i].xs && N[k].xe >= N[i].xe) {
					N[i].parent = k;
					N[k].children[N[k].nchild] = i;
					N[k].nchild++;
					ASSERT(N[k].nchild < MAX_C);
					break;
				}
			}
		}
		lastnz[j] = i;
	}
	// done with linking
	int firstPeak;
	//	Setup zone counter;
	m_nz = 0;
	if(N[0].xs > ileft) m_nz = 1;
	firstPeak = m_nz;
	//	Traverse Tree (Depth First)
	int *stack = new int[nz];
	int nst;
	int jt, fw, bk;
	int lf, rg;
	int lf0, rg0;
	for(k=0; k<nz && N[k].y==0; k++) {	// for each node in the lowest row
		nst = 0;
		stack[nst++] = k;	// push lowest row node into stack
		while(nst>0) {
			j = stack[--nst];	// pop node
			lf0 = lf = N[j].xs;
			fw = 1;
			while(N[j].nchild>0) {
				if(N[j].nchild > 1) {
					for(i=N[j].nchild-1; i>0; i--) stack[nst++] = N[j].children[i];
				}
				j = N[j].children[0];
				lf += N[j].xs;	// *************
				fw++;
			}
			//	correct value of left end
			//lf /= fw;	// *************
			lf = lf0;
			// backtrack to find right end
			bk = 1;
			rg0 = rg = N[j].xe;	// we need to initialize in case only one step back
			while(N[j].parent>-1) {
				jt = N[j].parent;
				if(N[jt].nchild>1 && j != N[jt].children[N[jt].nchild-1]) break;
				else j = jt;
				rg0 = N[j].xe;
				rg += N[j].xe;	// **************
				bk++;
			}
			//rg /= bk;	// *************
			rg = rg0;
			if( (fw>1 || bk>1) ) {	// we want to get rid of (1,1) paths that certainly are noise
				// create zone (lf, rg) with strength fw,bk
				m_zone[m_nz].i1 = lf;	m_zone[m_nz].i2 = rg;
				m_zone[m_nz].Peak[0]  = fw;
				m_zone[m_nz].Peak[1] = bk;
				m_nz += 2;
			}
		}
	}
	delete[] stack;
	delete[] N;

	// fill in missing zones (due to "deeps")
	if(firstPeak==1) {
		m_zone[0].i1 = ileft;	m_zone[0].i2 = m_zone[1].i1;
		for(i=0; i<2; i++) m_zone[0].Peak[i] = 0;
		k = 2;
	}
	else k = 1;
	for(; k<m_nz-2; k+=2) {
		m_zone[k].i1 = m_zone[k-1].i2;	m_zone[k].i2 = m_zone[k+1].i1;
		for(i=0; i<2; i++) m_zone[k].Peak[i] = 0;
	}
	if(m_zone[m_nz-2].i2 < iright) {
		m_zone[m_nz-1].i1 = m_zone[m_nz-2].i2;	m_zone[k].i2 = iright;
		for(i=0; i<2; i++) m_zone[m_nz-1].Peak[i] = 0;
	}
	else m_nz--;
	// missing zones filled

	// merge zones with a narrow gap - Infrequently used
	int min_gap = 2 + m_min_level;	// needs to be computed - H_MIN_INTERVAL/2 ?
	// minimum depth of peaks (at least on one side)
	int min_depth = m_min_level;	// m_min_level should be at least 1; if 0, then the following step is ineffectual
	if(forced>0) {
		for(k = firstPeak; k<m_nz; k+= 2) {
			if(m_zone[k].Peak[0] < min_depth && m_zone[k].Peak[1] < min_depth) {	// eliminate peak zone
				if(k==0) m_zone[k+1].i1 = m_zone[k].i1;	// peak is absorbed by next gap
				else {
					if(k==(m_nz-1)) m_zone[k-1].i2 = m_zone[k].i2; // peak is absorbed by previous gap
					else {// peak and next gap are absorbed by next gap
						m_zone[k+1].i1 = m_zone[k-1].i1;	// we keep prev. eliminate current and next
						m_zone[k-1].i2 = m_zone[k-1].i1;	// flag gap zone for elimination
					}
				}
				m_zone[k].i2 = m_zone[k].i1;	// flag peak zone for elimination
			}
		}
		ZoneCleanUp();
	}
	else {
		for(k=firstPeak; k<m_nz-2; k+= 2) {
			if(m_zone[k+2].i1 - m_zone[k].i2 > min_gap) continue;
			if(m_zone[k].Peak[1] < min_depth && m_zone[k+2].Peak[0] < min_depth) {
				m_zone[k+2].i1 = m_zone[k].i1;	m_zone[k+2].Peak[0] = m_zone[k].Peak[0];	// second absurbs previous
				m_zone[k].i2 = m_zone[k].i1;	// flag for elimination
				m_zone[k+1].i2 = m_zone[k+1].i1; // flag gap zone for elimination - that's the gap we are elminating
			}
		}
		ZoneCleanUp();
	}

	// eliminate narrow peaks (i.e. peaks with too few values) - Infrequently used
	// but not when they have a lot of values!
	for(k = firstPeak; k<m_nz-2; k+= 2) {
		if(m_zone[k].i2 - m_zone[k].i1 > min_gap) continue;
		if(m_zone[k].Peak[0] <= min_depth && m_zone[k].Peak[1] <= min_depth) {	// merge zones
			m_zone[k].i2 = m_zone[k].i1;	// flag for elimination
			if(k>0) m_zone[k+1].i1 = m_zone[k-1].i1;
			else m_zone[k+1].i1 = m_zone[k].i1;
			if(k>0) m_zone[k-1].i2 = m_zone[k-1].i1; // flag gap zone for elimination
		}
	}
	ZoneCleanUp();

	for(k=0; k<m_nz; k++) m_zone[k].Stats(m_hh);	// a bit wasteful, but OK for now

	// eliminate peaks with too few pixels - we should do better with second order stats
	double too_few_pixels = color==1? 0.2: 1.0;	// perecent of total - this is rather arbitrary, see above
	for(k = firstPeak; k<m_nz-2; k+= 2) {
		if(m_zone[k].sum < too_few_pixels) {
			// next 4 statements are repeated code from above
			m_zone[k].i2 = m_zone[k].i1;	// flag for elimination
			if(k>0) m_zone[k+1].i1 = m_zone[k-1].i1;
			else m_zone[k+1].i1 = m_zone[k].i1;
			if(k>0) m_zone[k-1].i2 = m_zone[k-1].i1; // flag gap zone for elimination
			m_zone[k+1].Stats(m_hh);
		}
	}
	ZoneCleanUp();

	if(m_nz>1) {
		AbsorbEndZone(LEFT_Z);
		AbsorbEndZone(RIGHT_Z);
	}
	if(m_parcelZones==TRUE) EliminateSparseZones(1);
	else EliminateSparseZones(0);	// for better stats !!!

	return 1;
}