/*
* Copyright 2003(c) Theo Pavlidis
* See pack4png.h for a detailed notice on the use of the software
*/	

// HistAnalArray.h: interface for the CHistAnalArray class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_HISTANALARRAY_H__F61252FA_02F2_4985_ABE1_DAB38DE07CD0__INCLUDED_)
#define AFX_HISTANALARRAY_H__F61252FA_02F2_4985_ABE1_DAB38DE07CD0__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#define LEFT_Z	0
#define RIGHT_Z	1

#define MR_LVL	9

//	Parameters that are fixed for a normalized histogram on [0,255]
#define H_MEAN_MAX_GAP		5	// difference between mean and max location (0 if perfect Gaussian)
#define H_MIN_INTERVAL		6	// should be at least 2 to take care of noise - broken scanner bit
								// the value is actually the sum of two adjacent intervals
#define H_LARGE_AMPLITUDE	2	// used in conjuction with min_interval to negate
								// elimination of ampl. around transition is large. value 1 cause anomalies

#define H_RARE_VALUES		5	// sum of normalized (percent) intensities in a zone below this
								// value is assumed to represent rare intensities
#include <math.h>

typedef struct _hZone {
	int		i1, i2;
	int		Peak[2];
	//int		Peakx;
	int		imax;
	double	mean;
	double	sum;
	void display(CString *sp) {
		if(imax<0) sp->Format("[%03d,%03d) empty", i1, i2);
		else sp->Format("[%03d,%03d) pk=%d,%d sum=%5.2f mean=%5.2f imax=%d",
			i1, i2, Peak[0], Peak[1], sum, mean, imax);
	}
	void copyto(struct _hZone *zp) {
		zp->i1 = i1;	zp->i2 = i2;		zp->Peak[0] = Peak[0];	zp->Peak[1] = Peak[1];
		zp->sum = sum;	zp->imax = imax;	zp->mean = mean;
	}
	BOOL SpreadOut()
	{
		if(sum < H_RARE_VALUES) return TRUE;
		if(fabs(mean - imax) > H_MEAN_MAX_GAP) return TRUE;
		return FALSE;
	}
	void Stats(double hh[])
	{
		int i;
		double maxval = hh[i1];
		imax = i1;
		for(i=i1+1; i<i2; i++) {
			if(hh[i] > maxval) { maxval = hh[i]; imax = i; }
		}
		sum = mean = 0;
		for(i=i1; i<i2; i++) {
			sum  += hh[i];
			mean += hh[i]*i;
		}
		if(sum>0) mean /= sum;
	}
} hZone;

class CHistAnalArray  
{
protected:
private:
	long	*m_hi;		// original histogram	[directly from INPUT]
	double	*m_hh;		// normalized histogram	[directly from INPUT]
	int		m_lvl;		// dimension of above two arrays	[directly from INPUT]
	hZone	*m_zone;	// the output of the program
	int		m_nz;		// actual number of zones found
	double  m_err;		//integral square error for color of histogram
	BOOL	m_parcelZones;	// whether to split weak zones 	[directly from INPUT]
	int		m_min_level;// main variable parameter
public:
	CHistAnalArray();
	virtual ~CHistAnalArray();
	BOOL Initialize(long hist[], int levels, int strategy, BOOL answer);	// new version 4/11/03
	int  FormZones(int forced, int color);
	void ListZones(CString note);	// in debugging mode
	int  Assign(int pxmap[], int mode, RECT z[]);
	double error() { return m_err; }

protected:
	void AbsorbEndZone (int n);
	void EliminateSparseZones(int force);
	BOOL MakeZoneSymmetric(BOOL first);
	void ZoneCleanUp();
};


#endif // !defined(AFX_HISTANALARRAY_H__F61252FA_02F2_4985_ABE1_DAB38DE07CD0__INCLUDED_)
