
/*
 * Functions for modifying an image so that it can be encoded more efficiently
 * with PNG.
 *
 * COPYRIGHT NOTICE, DISCLAIMER, and LICENSE:

 * Copyright 2003(c) Theo Pavlidis

 * This is 1.1 version (July 2003) of the library and the USE AT YOUR OWN RISK
 * warning remains applicable.
 *
 * Permission is hereby granted to use, copy, modify, and distribute this
 * source code, or portions hereof, for any purpose, without fee, subject
 * to the following restrictions:
 *
 * 1. The origin of this source code must not be misrepresented.
 *
 * 2. Altered versions must be plainly marked as such and
 * must not be misrepresented as being the original source.
 *
 * 3. This Copyright notice may not be removed or altered from
 *    any source or altered source distribution.

 * Comments are welcome. Send to t.pavlidis@ieee.org

*/

#ifndef PACK4PNG_H
#define PACK4PNG_H

#ifndef _WINDEF_	// for porting outside MS Windows
typedef int                 BOOL;
typedef unsigned char       BYTE;
typedef long				LONG;

typedef struct _RECT {
   LONG left;
   LONG top;
   LONG right;
   LONG bottom;
} RECT;
#endif //_WINDEF_

#define COLOR_P4P	1000
#define FORCED_P4P	 100

#ifdef P4P_DLL_BUILD
#define P4P_DLL_FCN __declspec(dllexport)
#else
#define P4P_DLL_FCN __declspec(dllimport)
#endif

extern "C" P4P_DLL_FCN int pack4png0(
			long hist[],	// input: histogram
			int compression,// min_level of histogram crossings, etc
			int pxmap[]		// output: array giving new pixel values in terms of old
			);

extern "C" P4P_DLL_FCN int pack4png1(
			long hist[],	// input: histogram
			int levels,		// input: usually 256
			int compression,// min_level of histogram crossings, etc
			int pxmap[],	// output: array giving new pixel values in terms of old
			double *erp		// intergal square error pointer
			);

extern "C" P4P_DLL_FCN int pack4png2(
			long hist[],	// input: histogram
			int levels,		// input: usually 256
			int compression,// min_level of histogram crossings, etc
			BOOL answer,	// if we want to know details
			int pxmap[],	// output: array giving new pixel values in terms of old
			double *erp,		// intergal square error pointer
			RECT z[]		// info about zones
			);

extern "C" P4P_DLL_FCN int pack4gct0(
			long hist[],	// input: histogram
			int compression,// min_level of histogram crossings, etc
			int pxmap[]		// output: array giving new pixel values in terms of old
			);

extern "C" P4P_DLL_FCN int pack4gct1(
			long hist[],	// input: histogram
			int levels,		// input: usually 256
			int compression,// min_level of histogram crossings, etc
			int pxmap[],	// output: array giving new pixel values in terms of old
			double *erp		// intergal square error pointer
			);

extern "C" P4P_DLL_FCN int pack4gct2(
			long hist[],	// input: histogram
			int levels,		// input: usually 256
			int compression,// min_level of histogram crossings, etc
			BOOL answer,	// if we want to know details
			int pxmap[],	// output: array giving new pixel values in terms of old
			double *erp,		// intergal square error pointer
			RECT z[]		// info about zones
			);

#endif //PACK4PNG_H
